{-------------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.1 (the "License"); you may not use this file except in compliance
  with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
  the specific language governing rights and limitations under the License.

  The Original Code is: HexView.pas, Peter Schraut (www.console-dev.de), 
  released 2004-06-22.

  Alternatively, the contents of this file may be used under the terms of the
  GNU General Public License Version 2 or later (the "GPL"), in which case
  the provisions of the GPL are applicable instead of those above.
  If you wish to allow use of your version of this file only under the terms
  of the GPL and not to allow others to use your version of this file
  under the MPL, indicate your decision by deleting the provisions above and
  replace them with the notice and other provisions required by the GPL.
  If you do not delete the provisions above, a recipient may use your version
  of this file under either the MPL or the GPL.


  13th August 2004
    + MouseWheel Support
    + Drag&Drop for files
    
-------------------------------------------------------------------------------}
unit HexView;

interface
uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, ComCtrls, ShellAPI;

type
  TCustomHexEdit = class;
  TCustomHexGutter = class;

  THexGutterPaintInfo = record
    nChars: Integer;
    nRows: Integer;
    StartOffset: Int64;
  end;
  
  TCustomHexGutter = class(TPersistent)
  private
    FAutoSize: Boolean;
    FColor: TColor;
    FCursor: TCursor;
    FDecimal: Boolean;
    FDigitCount: Integer;
    FFont: TFont;
    FLeadingZero: Boolean;
    FMarginX: Integer;
    FOnChange: TNotifyEvent;
    FOwner: TCustomHexEdit;
    FRealWidth: Integer;
    FWidth: Integer;    
    procedure OnFontChange(Sender: TObject);
    procedure SetColor(NewColor: TColor);
    procedure SetFont(NewFont: TFont);
  protected
    procedure Change; virtual;
    function ClientRect: TRect;
    function GetLineText(Offset: Int64): string; virtual;
    procedure Paint(const TargetCanvas: TCanvas; const PaintInfo: THexGutterPaintInfo); virtual;
    property Color: TColor read FColor write SetColor;
    property Cursor: TCursor read FCursor write FCursor;
    property Font: TFont read FFont write SetFont;    
    property OnChange: TNotifyEvent read FOnChange write FOnChange;    
  public
    constructor Create(AOwner: TCustomHexEdit);
    destructor Destroy; override;
  end;
  
  
  THexGutter = class(TCustomHexGutter)
  published
    property Color;
    property Cursor;
    property Font;
  end;
  

  THexScrollBar = class(TScrollBar)
  private
    function GetHexOwner: TCustomHexEdit;
  protected
    procedure Scroll(ScrollCode: TScrollCode; var ScrollPos: Integer); override;
  public
    constructor Create(AOwner: TCustomHexEdit); reintroduce;
    property Owner: TCustomHexEdit read GetHexOwner;
  end;
    
    
  TEditorOptions = set of
  (
    //eoAlwaysShowCaret,
    //eoNoCaret,
    eoAcceptDropFiles
  );
  
  
  TCustomHexEdit = class(TCustomControl)
  private
    FBackBuffer: TBitmap;
    FByteBuffer: array [0..4096] of Byte;
    FBytesPerLine: Integer;
    FCaretPos: Int64;
    FColumnSpace: Integer;
    FGutter: THexGutter;
    FOptions: TEditorOptions;
    FStream: TStream;
    FTopLine: Int64;
    FVertScrollBar: THexScrollBar;
    procedure SetAcceptDropFiles(Value: Boolean);
    procedure SetOptions(Value: TEditorOptions); 
    procedure SetStream(S: TStream);
    procedure WMDropFiles(var Message: TWMDropFiles); message WM_DROPFILES;
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure WMKillFocus(var Message: TWMSetFocus); message WM_KILLFOCUS;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMVScroll(var Message: TWMVScroll); message WM_VSCROLL;
  protected
    procedure Click; override;
    function DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): Boolean; override;
    function DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): Boolean; override;
    function GetTopLine: Int64;
    procedure HideCaret; virtual;
    procedure InitializeCaret; virtual;
    function IsPointInGutterArea(Pt: TPoint): Boolean; virtual;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure Paint; override;
    procedure PaintHex; virtual;
    procedure Resize; override;
    property BytesPerLine: Integer read FBytesPerLine write FBytesPerLine;
    property Stream: TStream read FStream write SetStream;
    property VertScrollBar: THexScrollBar read FVertScrollBar;
  public
    constructor Create(AOwner: TComponent); override;  
    destructor Destroy; override;
    procedure CloseFile;
    function IsShowingFile: Boolean;
    procedure OpenFile(AFilename: TFilename);
    procedure UpdateCaret; virtual;
    procedure SetFocus; override;
    property Options: TEditorOptions read FOptions write SetOptions;
    property TopLine: Int64 read FTopLine write FTopLine;

    
  end;

  
  THexEdit = class(TCustomHexEdit)
  published
    property BytesPerLine;
  end;
  
  
implementation

uses 
  Types;

var
  ByteToPrettyChar: array[0..255] of Char;
  ByteStringLUT: array [0..255] of string;
  
  
{ TCustomHexGutter }
constructor TCustomHexGutter.Create(AOwner: TCustomHexEdit);
begin
  inherited Create; 
  
  FOwner := AOwner;
  FOnChange := nil;
  
  FFont := TFont.Create;
  with FFont do
  begin
    Name := 'Courier New';
    Size := 10;
    Style := [];
    OnChange := OnFontChange;
  end;
  
  FAutoSize := True;
  FColor := clBtnFace;
  FCursor := crArrow; // crHandPoint;
  FDecimal := False;
  FDigitCount := 8;
  FLeadingZero := True;
  FMarginX:= 6;
  FWidth := 30;
end;

destructor TCustomHexGutter.Destroy;
begin       
  FOwner := nil;
  FOnChange := nil;
  FFont.Free;
  
  inherited;
end;

procedure TCustomHexGutter.Change;
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

function TCustomHexGutter.ClientRect: TRect;
begin
  Result := Rect(0, 0, FRealWidth, FOwner.Height);
end;

function TCustomHexGutter.GetLineText(Offset: Int64): string;
const
  cDecOrHex: array [Boolean] of string  = ('%x','%d');
  cZeroOrSpace: array [Boolean] of Char = (' ', '0');
begin
  Result := Format(cDecOrHex[FDecimal] ,[Offset]);    

  if (not FAutoSize) or FLeadingZero then
    while Length(Result) < FDigitCount do
      Result := cZeroOrSpace[FLeadingZero] + Result;
end;

procedure TCustomHexGutter.OnFontChange(Sender: TObject);
begin
  Change;
end;

procedure TCustomHexGutter.Paint(const TargetCanvas: TCanvas; const PaintInfo: THexGutterPaintInfo);
// paint is called by THexEdit, the passed TargetCanvas is the backbuffer.
var
  H: Integer;
  S: string;
  Y: Integer;
  Rows: Integer;
begin
  with TargetCanvas do
  begin
    Font := Self.Font;
    Brush.Color := Self.Color; 

    H := ClipRect.Bottom-ClipRect.Top;
    Rows := PaintInfo.nRows;

    S := GetLineText(PaintInfo.StartOffset+Rows);  
    if FAutoSize then
      FRealWidth := TextWidth(S) + FMarginX
    else
    begin
      FRealWidth := FWidth;
      if FRealWidth < TextWidth(S)+FMarginX then
        FRealWidth := TextWidth(S) + FMarginX;
    end;

    FillRect(Rect(0, 0, FRealWidth, H));

    Pen.Color := clWindow;
    PenPos := Point(FRealWidth-2, 0);
    LineTo(FRealWidth-2, H);  

    for Y := 0 to Rows do
    begin
      S := GetLineText(PaintInfo.StartOffset+(Y*PaintInfo.nChars));       
      TextOut(FRealWidth - TextWidth(S) - 4, Y * TextHeight('Wg'), S);
    end;
  end;
end;

procedure TCustomHexGutter.SetColor(NewColor: TColor);
begin
  if FColor <> NewColor then
  begin
    FColor := NewColor;
    Change;
  end;
end;

procedure TCustomHexGutter.SetFont(NewFont: TFont);
begin
  if FFont <> NewFont then
  begin
    FFont.Assign(NewFont);
    Change;
  end;  
end;


{ THexScrollBar }
constructor THexScrollBar.Create(AOwner: TCustomHexEdit);
begin
  Assert(AOwner is TCustomHexEdit, 'AOwner must be TCustomHExEdit');
  
  inherited Create(AOwner);
end;

function THexScrollBar.GetHexOwner: TCustomHexEdit;
begin
  Result := (inherited Owner) as TCustomHexEdit;
end;

procedure THexScrollBar.Scroll(ScrollCode: TScrollCode; var ScrollPos: Integer);
begin
  inherited;

  if Assigned(Owner) and Assigned(Owner.Stream) then
  begin
    Owner.Stream.Position := ScrollPos * Owner.BytesPerLine;
    Max := Owner.Stream.Size div Owner.BytesPerLine;
    LargeChange := Owner.BytesPerLine * 1024;
    
    if Owner.Stream.Position > Owner.Stream.Size-256 then
      Owner.Stream.Position := Owner.Stream.Size-256;
  end;
end;


{ TCustomHexEdit }
constructor TCustomHexEdit.Create(AOwner: TComponent);
begin
  FBackBuffer := TBitmap.Create;
  FColumnSpace := 8;
  FGutter := THexGutter.Create(Self);
  FTopLine := 0;
  FBytesPerLine := 24;
  
  inherited;
  
  Font.Name := 'Courier New';
  Font.Size := 10;
  Font.Style:= [];

  FVertScrollBar := THexScrollBar.Create(Self);
  FVertScrollBar.Kind := sbVertical;
  FVertScrollBar.Parent := Self;
  FVertScrollBar.Align := alRight;
  
  Color := clWindow;

  Options := []; 
end;

destructor TCustomHexEdit.Destroy;
begin
  CloseFile;
  FVertScrollBar.Free;
  FBackBuffer.Free;
  FGutter.Free;
  
  inherited;
end;


procedure TCustomHexEdit.Click;
begin
  inherited;
  SetFocus;
end;

function TCustomHexEdit.DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): Boolean; 
begin
  Result := inherited DoMouseWheelDown(Shift, MousePos);

  
  VertScrollBar.Position := VertScrollBar.Position + 1;
  FStream.Position := VertScrollBar.Position * FBytesPerLine;
  Refresh;
end;

function TCustomHexEdit.DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): Boolean;    
begin
  Result := inherited DoMouseWheelUp(Shift, MousePos);

  VertScrollBar.Position := VertScrollBar.Position - 1;
  FStream.Position := VertScrollBar.Position * FBytesPerLine;
  Refresh;
end;

function TCustomHexEdit.GetTopLine: Int64;
begin
  if Assigned(FStream) then
    Result := FStream.Position div FBytesPerLine
  else
    Result := 0;
end;

procedure TCustomHexEdit.HideCaret;
begin
  Windows.HideCaret(Handle);
end;

procedure TCustomHexEdit.InitializeCaret; 
begin       {
  if Focused and not (eoNoCaret in FOptions) then
  begin
    CreateCaret(Handle, 0, Canvas.TextWidth('W'), Canvas.TextHeight('Wg'));
    UpdateCaret;
  end;   }
end;

function TCustomHexEdit.IsPointInGutterArea(Pt: TPoint): Boolean;
begin
  Result := PtInRect(FGutter.ClientRect, Pt);
end;

function TCustomHexEdit.IsShowingFile: Boolean;
begin
  Result := Assigned(FStream) and (FStream.Size > 0);
end;

procedure TCustomHexEdit.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;

  if IsPointInGutterArea(Point(X,Y)) then
    Cursor := FGutter.Cursor
  else
    Cursor := crDefault;
end;

procedure TCustomHexEdit.OpenFile(AFilename: TFilename);
begin
  CloseFile;
  
  FStream := TFileStream.Create(AFilename, fmOpenRead or fmShareDenyWrite);
  VertScrollBar.Position := 0;
  Resize;
end;

procedure TCustomHexEdit.CloseFile;
begin
  if Assigned(FStream) then FreeAndNil(FStream);
  Refresh;
end;

procedure TCustomHexEdit.Paint;
var
  GutterPaintInfo: THexGutterPaintInfo;
begin
  FBackBuffer.Canvas.Brush.Color := clWindow; // Color;
  FBackBuffer.Canvas.Font := Self.Font;
  
  with FBackBuffer.Canvas do
  begin
    // clear
    FillRect(ClientRect);

    // paint the gutter
    GutterPaintInfo.nChars := FBytesPerLine;
    GutterPaintInfo.nRows := ClientHeight div Canvas.TextHeight('Wg');
    GutterPaintInfo.StartOffset := GetTopLine * FBytesPerLine;
    FGutter.Paint(FBackBuffer.Canvas, GutterPaintInfo);

    FBackBuffer.Canvas.Font := Self.Font;
    PaintHex;
  end;
   
  // swap backbuffer
  BitBlt(Canvas.Handle, 0, 0, ClientWidth-FVertScrollBar.Width-1, ClientHeight, FBackBuffer.Canvas.Handle, 0, 0, SRCCOPY); 

  UpdateCaret;  
end;

procedure TCustomHexEdit.PaintHex;
const
  cHighlightBkColors: array [Boolean] of TColor = (clWindow, $fff5f5);
  cHighlightFontColors: array [Boolean] of TColor = (clWindowText, clWindowText);
var
  X,Y: Integer;
  
  nRead: Integer;
  CharW,
  CharH: Integer;
  nCharsPerLine: Integer;
  CurX: Integer;

  procedure PaintDoubleLine(X0, Y0, X1, Y1, DistanceX, DistanceY: Integer; AColor: TColor);
  begin
    with FBackBuffer.Canvas do
    begin    
      Pen.Color := AColor;
      PenPos := Point(X0, Y0);
      LineTo(X1, Y1);
          
      Pen.Color := AColor;
      PenPos := Point(X0+DistanceX, Y0+DistanceY);
      LineTo(X1+DistanceX, Y1+DistanceY);
    end;
  end;
  
begin
  
  if not Assigned(FStream) then Exit;
  
  with FBackBuffer.Canvas do
  begin  

    CharW := TextWidth('W');
    CharH := TextHeight('Wg');

    FBytesPerLine := (ClientWidth-FGutter.FRealWidth) div (CharW*2+FColumnSpace);
    Dec(FBytesPerLine, FBytesPerLine div 3);
      
    nCharsPerLine := FBytesPerLine;
      
    for Y := 0 to (ClientHeight div CharH) do
    begin
      nRead := FStream.Read(FByteBuffer, nCharsPerLine);
    
      if nRead > 0 then
      begin
        CurX := FGutter.FRealWidth + FColumnSpace;
          
        for X := 0 to nRead-1 do
        begin
          Brush.Color := cHighlightBkColors[(X and 1)=1];
          Font.Color  := cHighlightFontColors[(X and 1)=1];        

          TextOut(CurX, Y*CharH, ByteStringLUT[FByteBuffer[X]]);
          Inc(CurX, CharW*2+FColumnSpace);
        end;

        CurX := nCharsPerLine * (CharW*2+(FColumnSpace)) + nCharsPerLine;
        Inc(CurX, FGutter.FRealWidth+FColumnSpace);

        PaintDoubleLine(CurX-5, 0, CurX-5, Height, 2, 0, clMedGray);

        Inc(CurX, FColumnSpace);          
        for X := 0 to nRead-1 do
        begin
          Brush.Color := clWindow;
          Font.Color  := clWindowText;        

          TextOut(CurX, Y*CharH, ByteToPrettyChar[FByteBuffer[X]]);
          Inc(CurX, CharW);
        end;    

        CurX := nCharsPerLine * (CharW*3+(FColumnSpace)) + nCharsPerLine;
        Inc(CurX, FGutter.FRealWidth+FColumnSpace*3);
        PaintDoubleLine(CurX+5, 0, CurX+5, Height, 2, 0, clMedGray);          
      end
      else
        Break;
    end;
  end;
end;

procedure TCustomHexEdit.Resize;
var
  VertBarPos: Integer;
begin
  VertBarPos := FVertScrollBar.Position;
  
  inherited;
  
  FBackBuffer.Width := ClientWidth;
  FBackBuffer.Height:= ClientHeight;
  FVertScrollBar.Scroll(scTrack, VertBarPos);
  Refresh;
end;

procedure TCustomHexEdit.SetFocus;
begin
  inherited;
end;

procedure TCustomHexEdit.SetAcceptDropFiles(Value: Boolean);
begin
  if not (csDesigning in ComponentState) then
    DragAcceptFiles(Handle, LongBool(Value));
end;

procedure TCustomHexEdit.SetOptions(Value: TEditorOptions);
begin
  if FOptions <> Value then
  begin
    SetAcceptDropFiles(eoAcceptDropFiles in Value);

    FOptions := Value;
  end;
end;

procedure TCustomHexEdit.SetStream(S: TStream);
begin
  if FStream <> S then
  begin
    FStream := S;

    FVertScrollBar.Enabled := Assigned(FStream);    
    Refresh;
  end;
end;

procedure TCustomHexEdit.UpdateCaret;
begin             {
  if not (eoNoCaret in FOptions) then
  begin    
    //SetCaretPos(72,103); 
    //ShowCaret(Handle);
  end;     }
end;

procedure TCustomHexEdit.WMDropFiles(var Message: TWMDropFiles);
var
  DropName: string;
begin
  inherited;

  if DragQueryFile(Message.Drop, $ffffffff, nil, 0) > 0 then
  begin
    SetLength(DropName, DragQueryFile(Message.Drop, 0, nil, 0)+1);
    DragQueryFile(Message.Drop, 0, PChar(DropName), Length(DropName));
    DragFinish(Message.Drop);
    

    if FileExists(DropName) then
    begin
      OpenFile(DropName);
      Message.Result := 0;
    end;

    SetLength(DropName, 0);
  end; 
end;

procedure TCustomHexEdit.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  Message.Result := 1;
end;

procedure TCustomHexEdit.WMKillFocus(var Message: TWMSetFocus);
begin
  inherited;
             {
  if not (eoAlwaysShowCaret in FOptions) then
    HideCaret;   }
end;

procedure TCustomHexEdit.WMSetFocus(var Message: TWMSetFocus);
begin
  inherited;
  InitializeCaret;
end;

procedure TCustomHexEdit.WMVScroll(var Message: TWMVScroll);
begin
  inherited;

  if Visible then
    SetFocus;
       
  if Assigned(Stream) then
  begin
    TopLine := Stream.Position;   
    Refresh;
  end;        
end;

{ initialization }
procedure DoInitialization;
var
  I: Integer;
begin
  for I := 0 to 255 do
  begin
    ByteStringLUT[I] := Format('%.2x', [I]);
    ByteToPrettyChar[I] := Char(I);
    
    if I < 32 then ByteToPrettyChar[I] := '.';
  end;
end;
  
initialization
  DoInitialization;

  
end.
